#  tests for pygeos-0.9-py37h66a1372_1 (this is a generated file);
print('===== testing package: pygeos-0.9-py37h66a1372_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import numpy as np
from pygeos import box, area, intersection

polygons_x = box(range(5), 0, range(10, 15), 10)
polygons_y = box(0, range(5), 10, range(10, 15))

area(intersection(polygons_x[:, np.newaxis], polygons_y[np.newaxis, :]))

#  --- run_test.py (end) ---

print('===== pygeos-0.9-py37h66a1372_1 OK =====');
print("import: 'pygeos'")
import pygeos

